"""Represents a disy Cadenza analytics extension that is accessible via disy Cadenza at entry points
 depending on the extension type. The extension will run its main analytics function when invoked
 with a POST on the relative path."""
import json
from io import StringIO
from typing import Callable, List

import pandas as pd
from flask import Response, request

from cadenzaanalytics.data.analytics_extension import AnalyticsExtension
from cadenzaanalytics.data.attribute_group import AttributeGroup
from cadenzaanalytics.data.extension_type import ExtensionType
from cadenzaanalytics.data.parameter import Parameter
from cadenzaanalytics.request.analytics_request import AnalyticsRequest
from cadenzaanalytics.request.request_metadata import RequestMetadata
from cadenzaanalytics.response.extension_response import ExtensionResponse


class CadenzaAnalyticsExtension:
    def __init__(self,
                 relative_path: str,
                 analytics_function: Callable[[RequestMetadata, pd.DataFrame], ExtensionResponse],
                 print_name: str,
                 extension_type: ExtensionType,
                 attribute_groups: List[AttributeGroup],
                 parameters: List[Parameter] = None):
        self._relative_path = relative_path
        self._analytics_function = analytics_function

        self._analytics_extension = AnalyticsExtension(print_name, extension_type, attribute_groups, parameters)

    @property
    def relative_path(self) -> str:
        return self._relative_path

    @property
    def print_name(self) -> str:
        return self._analytics_extension.print_name

    @property
    def extension_type(self) -> ExtensionType:
        return self._analytics_extension.extension_type

    def handle_request(self) -> Response:
        analytics_request = self._get_request_data(request)

        analytics_response = self._analytics_function(analytics_request.metadata, analytics_request.data)

        return analytics_response.get_response(analytics_request.metadata.get_all_columns(), analytics_request.data)

    def get_capabilities(self) -> Response:
        return Response(response=self._analytics_extension.to_json(), status=200, mimetype="application/json")

    def _get_request_data(self, multipart_request) -> AnalyticsRequest:
        metadata_dict = json.loads(multipart_request.form['metadata'])
        metadata = RequestMetadata(metadata_dict)

        if metadata.has_columns():
            type_mapping = {}
            for column in metadata.get_all_columns():
                type_mapping[column.name] = column.data_type.pandas_type()
            csv_data = StringIO(multipart_request.form['data'])
            df_data = pd.read_csv(csv_data, sep=";", dtype=type_mapping)
        else:
            df_data = pd.DataFrame()

        return AnalyticsRequest(metadata, df_data)



import json


# pylint: disable=protected-access
class DataObject:
    _attribute_mapping = {}
    _attribute_constructors = {}  # required for enums that are deserialized

    def _to_dict(self) -> dict:
        result = {}

        for key, attribute in self._attribute_mapping.items():
            value = getattr(self, attribute)

            if value is not None:
                if isinstance(value, list):
                    result_list = []

                    for element in value:
                        if issubclass(type(element), DataObject):
                            result_list.append(element._to_dict())
                        else:
                            result_list.append(element)

                    result[key] = result_list
                else:
                    result[key] = value

        return result

    def to_json(self, indent=None) -> str:
        return json.dumps(self._to_dict(), indent=indent, default=str)

    @classmethod
    def _from_dict(cls, data: dict):
        constructor_parameters = {}

        # remap data fields to constructor parameters
        for key, value in data.items():
            if key in cls._attribute_mapping:
                # TODO: Consider removing the underscore from attribute mapping, and adding it when used
                # get the constructor parameter key be removing underscore from attribute mapping
                parameter_key = cls._attribute_mapping[key][1:]

                if key in cls._attribute_constructors:
                    value = cls._attribute_constructors[key](value)
                constructor_parameters[parameter_key] = value

        return cls(**constructor_parameters)



from enum import Enum


class MeasureAggregation(Enum):
    COUNT = "count"
    COUNT_DISTINCT = "countDistinct"
    SUM = "sum"
    AVERAGE = "average"
    MIN = "min"
    MAX = "max"

    def __str__(self):
        return self.value



from typing import Optional

from cadenzaanalytics.data.geometry_type import GeometryType
from cadenzaanalytics.data.attribute_role import AttributeRole
from cadenzaanalytics.data.data_object import DataObject
from cadenzaanalytics.data.data_type import DataType
from cadenzaanalytics.data.measure_aggregation import MeasureAggregation


# pylint: disable=too-many-instance-attributes
class ColumnMetadata(DataObject):
    _attribute_mapping = {
        "name": "_name",
        "printName": "_print_name",
        "attributeGroupName": "_attribute_group_name",
        "dataType": "_data_type",
        "role": "_role",
        "measureAggregation": "_measure_aggregation",
        "format": "_format",
        "geometryType": "_geometry_type"
    }
    _attribute_constructors = {
        "dataType": DataType,
        "role": AttributeRole,
        "measureAggregation": MeasureAggregation,
        "geometryType": GeometryType
    }

    # pylint: disable=redefined-builtin
    def __init__(self,
                 name: str,
                 print_name: str,
                 attribute_group_name: str,
                 data_type: DataType,
                 role: AttributeRole,
                 measure_aggregation: MeasureAggregation = None,
                 format: str = None,
                 geometry_type: GeometryType = None):
        self._name = name
        self._print_name = print_name
        self._attribute_group_name = attribute_group_name
        self._data_type = data_type
        self._role = role
        self._measure_aggregation = measure_aggregation
        self._format = format
        self._geometry_type = geometry_type

    @property
    def name(self) -> str:
        return self._name

    @property
    def print_name(self) -> str:
        return self._print_name

    @property
    def attribute_group_name(self) -> str:
        return self._attribute_group_name

    @property
    def data_type(self) -> DataType:
        return self._data_type

    @property
    def role(self) -> AttributeRole:
        return self._role

    @property
    def measure_aggregation(self) -> Optional[MeasureAggregation]:
        return self._measure_aggregation

    @property
    def format(self) -> str:
        return self._format

    @property
    def geometry_type(self) -> GeometryType:
        return self._geometry_type



from typing import List

from pandas import DataFrame

from cadenzaanalytics.data.column_metadata import ColumnMetadata
from cadenzaanalytics.response.extension_data_response import ExtensionDataResponse


class ImageResponse(ExtensionDataResponse):
    def __init__(self, image):
        content_type = 'image/png'
        super().__init__(content_type)

        self._image = image

    def get_response(self, original_column_metadata: List[ColumnMetadata], original_data: DataFrame):
        return self._create_response(self._image)



from cadenzaanalytics.request.request_metadata import RequestMetadata


class AnalyticsRequest:
    def __init__(self, metadata: RequestMetadata, data):
        self._metadata = metadata
        self._data = data

    @property
    def metadata(self) -> RequestMetadata:
        return self._metadata

    @property
    def data(self):
        return self._data



from typing import List

from cadenzaanalytics.data.data_object import DataObject
from cadenzaanalytics.data.parameter_type import ParameterType


class Parameter(DataObject):
    _attribute_mapping = {
        "name": "_name",
        "printName": "_print_name",
        "parameterType": "_parameter_type",
        "options": "_options",
        "required": "_required",
        "defaultValue": "_default_value"
    }

    def __init__(self,
                 name: str,
                 print_name: str,
                 parameter_type: ParameterType,
                 options: List[str] = None,
                 required: bool = False,
                 default_value: any = None):
        self._name = name
        self._print_name = print_name
        self._parameter_type = parameter_type
        self._options = options
        self._required = required
        self._default_value = default_value
